##' Takes a \code{mids} object, and produces a new object of class \code{mids}
##'
##' This function is based on the \code{mice} function's principle, which uses the MICE algorithm. \code{mice} 
##' allows to generate multiple imputations for incomplete multivariate data by Gibbs sampling. The algorithm
##' imputes an incomplete colum (the target column) by generating 'plausible' synthetic values given other columns in the data.
##' Each incomplete column must act as a target column, and has its own set of predictors. The default set of predictors
##' for a given target consists of all others colums in the data. For predictors that are incomplete themselves, the most recently
##' generated imputations are used to complete the predictors prior to imputation of the target column. 
##'
##' Built-in elementary imputation methods are: 
##' \code{pmm} Bayesian linear regression (numeric)
##' \code{norm} Predictive mean matching (numeric)
##' \code{norm.nob} Linear regression ignoring model error (numeric)
##' \code{mean} Unconditional mean imputation (numeric)
##' \code{2l.norm} Two-level normal imputation (numeric)
##' \code{logreg} Logistic regression (factor, 2 categories)
##' \code{polyreg} Polytomous (unordered) logistic regression (factor, >=2 categories)
##' \code{lda} Linear discriminant analysis (factor, >=2 categories)
##' \code{sample} Random sample from the observed values (any)
##'
##' This function enables to impute missing values under the hypothesis of MNAR data, for one or more variable(s). A
##' sensitivity analysis can be performed using the \code{mids} object returned by the function.
##'
##' It goes through 3 steps:
##'
##' - 1. Estimate the parameters of the imputation model under ignorable missing data hypothesis using the function \code{mice}.
##' The fitted imputation model depends on the type of the variable which contains missing values, i.e. bayesian linear regression
##' for numeric variables, logistic regression for binary variables, polytomous unordered logistic regression for categorical
##' variables.
##'
##' - 2. Modify the imputation model according to the explored scenario by specifying supplementary parameters as arguments for the \code{sens.mice} function.
##' Without fitting a mixture model, we appeal to its principle as proposed by Rubin. Indeed, the addition of this(these) supplementary parameter(s) allows to specify
##' that the distribution of the variable of interest is different among subjects with missing value and among subjects witouht missing value, 
##' conditionnally on all variables included in the imputation model. The direction and the size of this difference is expressed by the
##' supplementary parameter in the imputation model.
##'
##' - 3. Impute the missing data using the function \code{sens.mice}, resulting into a \code{mids} object that contains m newly imputed data sets under MNAR hypothesis.
##' For this step, missing values are imputed with the usual MICE algorithm, using the previously modified imputation model.
##'
##' The sensitivity analysis can then be performed on the returned \code{mids} object, which contains the newly imputed data, according to
##' the assumption that values for the variabl(s) imputed with the \code{sens.mice} function come from two different distributions 
##' (for responders and incomplete responders).
##'
##' Built-in elementary imputation methods are: 
##'
##'     - \code{pmm}: Predictive mean matching (numeric)
##'    
##'     - \code{norm}: Bayesian linear regression (numeric) 
##'
##'     - \code{logreg}: Logistic regression (factor, 2 categories) 
##'
##'     - \code{polyreg}: Polytomous (unordered) regression (factor, >= 2 categories)   
##'
##' \code{ListMethod} is a vector of strings with length \code{ncol(IM$data)}, specifying 
##' the column(s) in data which has (have) to be imputed with an imputation model accounting for MNAR data. 
##' For variables which have to be imputed with a modified imputation model, 
##' the method has to be specified as \code{"MyFunc"}.
##' For variables which do not have to be imputed with a different imputation
##' model, the method has to be be specified as \code{""}.
##'
##' \code{SupPar} is a vector of numbers, specifying the supplementary parameters 
##' to be added to the imputation model(s). For \code{logreg} and \code{polyreg} methods,  
##' the supplementary parameters are expressed as odds-ratios (correponding to the excess of risk 
##' to present the modality of interest for non responders as compared to responders). 
##' The value for the reference need not be specified.
##' For \code{pmm} and \code{norm} methods, the supplementary parameters
##' are the difference between the expected values in responders and non responders.
##'
##' @title Multivariate Imputation by Chained Equations (Iteration Step for sensitivity analysis).
##' @export
##' @param IM An object of class \code{mids}, typically produces by a previous call
##' to \code{mice()} or \code{mice.mids()}
##' @param ListMethod A vector of strings with length \code{ncol(IM$data)}, specifying 
##' the colum(s) in data which has (have) to be imputed with a different imputation model
##' @param SupPar A vector of numbers, specifying the supplementary parameters 
##' to be added to the imputation model(s)
##' 
##' @return Returns an object of class \code{mids} (multiply imputed data set) with
##' usual components
##' @author Noemie Resseguier, with contributions of Roch Giorgi, 
##' David Hajage, Yann De Rycke and Xavier Paoletti
##' @references 
##' Resseguier, N., Giorgi, R. and Paoletti, X. (submitted) \emph{How to perform a senstivity analysis exploring the impact
##' of missing not at random data under different sceanrios of non response mechanism with the R software.} 
##' 
##' Rubin, D.B. \emph{Multiple Imputation for Nonresponse in Surveys.} New York: John Wiley & Sons, 1987.
##'
##' van Buuren, S., Groothuis-Oudshoorn, K. \emph{MICE: Multivariate Imputation by Chained Equations in R.}
##'
##' @seealso \code{\link{mice}}, \code{\link{mids}}
##' @examples 
##'
##' ###### In a descriptive context : data set : popmis
##'
##' # Do multiple imputation 
##' data(popmis)
##' MYpopmis <- popmis[, c("popular", "sex", "texp", "teachpop")]
##' IM <- mice(MYpopmis, method=c("pmm", "", "", ""), seed=13, maxit=5, m=5)
##'
##' # Describe popular
##'
##' # Construction of the completed imputed data sets
##' temp1 <- complete(IM, action=1)
##' temp2 <- complete(IM, action=2)
##' temp3 <- complete(IM, action=3)
##' temp4 <- complete(IM, action=4)
##' temp5 <- complete(IM, action=5)
##'
##' # Description of the popularity score
##' library(mitools)
##' miData <- imputationList(list(temp1, temp2, temp3, temp4, temp5))
##' MEAN <- with(miData, mean(popular))
##' VAR <- with(miData, var(popular))
##' MIcombine(MEAN, VAR)
##'
##' # Imputation with a supplementary parameter -0.5 on the popular variable
##' IMHyp0.5 <- sens.mice(IM, ListMethod = c("MyFunc", "", "", ""), SupPar = c(-0.5))
##'
##' # Describe popular
##'
##' # Construction of the completed imputed data sets
##' temp1Hyp0.5 <- complete(IMHyp0.5, action=1)
##' temp2Hyp0.5 <- complete(IMHyp0.5, action=2)
##' temp3Hyp0.5 <- complete(IMHyp0.5, action=3)
##' temp4Hyp0.5 <- complete(IMHyp0.5, action=4)
##' temp5Hyp0.5 <- complete(IMHyp0.5, action=5)
##'
##' # Description of the popularity score
##' library(mitools)
##' miDataHyp0.5 <- imputationList(list(temp1Hyp0.5, temp2Hyp0.5, temp3Hyp0.5,
##' temp4Hyp0.5, temp5Hyp0.5))
##' MEANHyp0.5 <- with(miDataHyp0.5, mean(popular))
##' VARHyp0.5 <- with(miDataHyp0.5, var(popular))
##' MIcombine(MEANHyp0.5, VARHyp0.5)
##'
##' ###### In a causal context : data set CHAIN
##'
##' library(mi)
##' data(CHAIN)
##'
##' MyData <- CHAIN[apply(CHAIN, 1, function(x) !all(is.na(x))), ]                                                                                         
##' 
##' ### VL2cat = binary variable (<400 c/mL vs >= 400 c/mL)
##' MyData$VL2cat <- NA
##' MyData$VL2cat <- ifelse((MyData$h39b.W1 < log(400) & !is.na(MyData$h39b.W1)),
##' 0, MyData$VL2cat)
##' MyData$VL2cat <- ifelse((MyData$h39b.W1 >= log(400) & !is.na(MyData$h39b.W1)),
##' 1, MyData$VL2cat)
##'
##' # Do multiple imputation
##'
##' MyDataVL2cat <- MyData[, c("age.W1", "c28.W1", "pcs.W1", "mcs37.W1", "b05.W1",
##' "haartadhere.W1", "VL2cat")]
##' MyDataVL2cat$mcs37.W1 <- as.factor(MyDataVL2cat$mcs37.W1)
##' MyDataVL2cat$b05.W1 <- as.factor(MyDataVL2cat$b05.W1)
##' MyDataVL2cat$haartadhere.W1 <- as.factor(MyDataVL2cat$haartadhere.W1)
##' MyDataVL2cat$VL2cat <- as.factor(MyDataVL2cat$VL2cat)
##' MyDataVL2catMI <- mice(MyDataVL2cat, method = c("pmm", "pmm", "pmm", "logreg",
##' "polyreg", "polyreg", "logreg"), seed = 13)
##' 
##' # Construction of the completed imputed data sets
##' mi1VL2cat <- complete(MyDataVL2catMI, action=1)
##' mi2VL2cat <- complete(MyDataVL2catMI, action=2)
##' mi3VL2cat <- complete(MyDataVL2catMI, action=3)
##' mi4VL2cat <- complete(MyDataVL2catMI, action=4)
##' mi5VL2cat <- complete(MyDataVL2catMI, action=5)
##' miDataVL2cat <- imputationList(list(mi1VL2cat, mi2VL2cat, mi3VL2cat,
##' mi4VL2cat, mi5VL2cat))
##'
##' # Description of VL2cat and fit of the model
##' library(mitools)
##' tablesVL2cat <- with(miDataVL2cat, prop.table(table(VL2cat)))
##'
##' fitVL2cat <- with(miDataVL2cat, glm(mcs37.W1 ~ as.numeric(age.W1) +  
##' as.numeric(pcs.W1) + as.factor(haartadhere.W1) + as.factor(VL2cat) +                                                
##' as.numeric(c28.W1) + as.factor(b05.W1), family=binomial))                                              
##'                                                
##' coefsVL2cat <- MIextract(fitVL2cat, fun=coef)
##' varsVL2cat <- MIextract(fitVL2cat, fun=vcov)
##' resVL2cat <- summary(MIcombine(coefsVL2cat, varsVL2cat))
##' devfitVL2cat <- MIextract(fitVL2cat, fun=deviance)
##'
##' fitVL2catPVAL <- with(miDataVL2cat, glm(mcs37.W1 ~ as.numeric(age.W1) +
##' as.numeric(pcs.W1) + as.factor(haartadhere.W1) +  as.numeric(c28.W1) +                                                 
##' as.factor(b05.W1), family=binomial))                                                                                                    
##' devfitVL2catPVAL <- MIextract(fitVL2catPVAL, fun=deviance)
##' 
##' pvalVL2cat <- mean(1-pchisq(unlist(devfitVL2catPVAL)-unlist(devfitVL2cat), 1))
##'
##' # Imputation with a supplementary parameter 1.2 on the VL2cat variable
##' MyDataVL2catMI.SCEN1 <- sens.mice(MyDataVL2catMI, ListMethod = c("", "", "", 
##' "", "", "", "MyFunc"), SupPar = c(1.2))
##' 
##' # Construction of the completed imputed data sets
##' mi1.SCEN1VL2cat <- complete(MyDataVL2catMI.SCEN1, action=1)
##' mi2.SCEN1VL2cat <- complete(MyDataVL2catMI.SCEN1, action=2)
##' mi3.SCEN1VL2cat <- complete(MyDataVL2catMI.SCEN1, action=3)
##' mi4.SCEN1VL2cat <- complete(MyDataVL2catMI.SCEN1, action=4)
##' mi5.SCEN1VL2cat <- complete(MyDataVL2catMI.SCEN1, action=5)
##' miData.SCEN1VL2cat <- imputationList(list(mi1.SCEN1VL2cat, mi2.SCEN1VL2cat,
##' mi3.SCEN1VL2cat, mi4.SCEN1VL2cat, mi5.SCEN1VL2cat))
##'
##' # Description of VL2cat and fit of the model 
##' tablesSCEN1VL2cat <- with(miData.SCEN1VL2cat, prop.table(table(VL2cat)))
##' 
##' fitVL2cat.SCEN1 <- with(miData.SCEN1VL2cat, glm(mcs37.W1 ~ as.numeric(age.W1) + 
##' as.numeric(pcs.W1) + as.factor(haartadhere.W1) + as.factor(VL2cat) +
##' as.numeric(c28.W1) + as.factor(b05.W1), family=binomial))                                                           
##'                                                           
##' coefs.SCEN1VL2cat <- MIextract(fitVL2cat.SCEN1, fun=coef)
##' vars.SCEN1VL2cat <- MIextract(fitVL2cat.SCEN1, fun=vcov)
##' res.SCEN1VL2cat <- summary(MIcombine(coefs.SCEN1VL2cat, vars.SCEN1VL2cat))
##' devfitVL2cat.SCEN1 <- MIextract(fitVL2cat.SCEN1, fun=deviance)
##'
##' fitVL2catPVAL.SCEN1 <- with(miData.SCEN1VL2cat, glm(mcs37.W1 ~ as.numeric(age.W1) + 
##' as.numeric(pcs.W1) + as.factor(haartadhere.W1) + as.numeric(c28.W1) +
##' as.factor(b05.W1), family=binomial))                                                               
##'                                                                
##' devfitVL2catPVAL.SCEN1 <- MIextract(fitVL2catPVAL.SCEN1, fun=deviance)
##' 
##' pvalVL2cat.SCEN1 <- mean(1-pchisq(unlist(devfitVL2catPVAL.SCEN1)-
##' unlist(devfitVL2cat.SCEN1), 1))






sens.mice <- function(IM, ListMethod = ListMethod, SupPar = SupPar){
if(length(ListMethod) > length(names(IM$data))){
  stop("You have specified too much new methods to be applied.")
} 
if(length(ListMethod) < length(names(IM$data))){
  stop("You have not specified enough new methods to be applied.")
}
cpt <- 0
for(i in 1:length(ListMethod)){
  if(ListMethod[i]=="MyFunc"){
    if(IM$method[i] == "norm" | IM$method[i] == "pmm" | IM$method[i] == "logreg"){
    cpt <- cpt + 1
    }
    if(IM$method[i] == "polyreg"){
    cpt <- cpt + dim(table(IM$data[i]))-1
    }
  }
  if(ListMethod[i]==""){
    cpt <- cpt
  }
}
if(length(SupPar) > cpt){
  stop("You have specified too much supplementary parameters to be applied.")
} 
if(length(SupPar) < cpt){
  stop("You have not specified enough supplementary parameters to be applied.")
} 
for(i in 1:length(ListMethod)){
  if(ListMethod[i] != "MyFunc" & ListMethod[i] != ""){
    stop("Values available for ListMethod are ''MyFunc'' and '' ''.")
  }
  if(ListMethod[i]=="Myfunc" & (IM$method[i] =="norm.nob")){
    stop("norm.nob is not an available method for the function sens.mice.")
  }
  if(ListMethod[i]=="Myfunc" & (IM$method[i] =="mean")){
    stop("mean is not an available method for the function sens.mice.")
  }
  if(ListMethod[i]=="Myfunc" & (IM$method[i] =="2l.norm")){
    stop("2l.norm is not an available method for the function sens.mice.")
  }    
  if(ListMethod[i]=="Myfunc" & (IM$method[i] =="lda")){
    stop("lda is not an available method for the function sens.mice.")
  }
  if(ListMethod[i]=="Myfunc" & (IM$method[i] =="sample")){
    stop("sample is not an available method for the function sens.mice.")
  }
  if(ListMethod[i]=="Myfunc" & (IM$method[i] =="polr")){
    stop("polr is not an available method for the function sens.mice.")
  }
} 
j <- 0
cpt <- 0
IMinit <- IM
MyMethod <- IM$method
assign("MyMethod", MyMethod, envir = .GlobalEnv)
assign("SupPar", SupPar, envir = .GlobalEnv)
listvar <- names(IM$data)
SumPr <- matrix(NA, nrow=length(ListMethod[ListMethod=="MyFunc"]), ncol=3)
for(ii in 1:length(listvar)){
  if(ListMethod[ii]=="MyFunc"){
    j <- j + 1
    cpt <- cpt + 1  
    if(MyMethod[ii]=="pmm" | MyMethod[ii]=="norm" | MyMethod[ii]=="logreg"){
      SumPrtemp <- c(listvar[ii], MyMethod[ii], SupPar[j])
      cat(SumPrtemp)   
    }
    if(MyMethod[ii]=="logreg"){
      if(SupPar[j] < 0){
        stop("Value for odds ratio can't be negative.") 
      }
    }
    if(MyMethod[ii]=="polyreg"){
      tempSupPar <- c(SupPar[j : (j + dim(table(IM$data[ii]))-2)])
      SumPrtemp <- c(listvar[ii], MyMethod[ii], tempSupPar)
      cat(SumPrtemp)
      for(m in 1:length(tempSupPar)){
        if(tempSupPar[m] < 0){
          stop("Value for odds ratio can't be negative.")  
        }
      }
    } 
    IMtemp <- IM
    IMtemp$pad$method <- c(rep("", length(names(IM$data))), rep("dummy", length(IM$pad$method) - length(names(IM$data))))
    IMtemp$pad$method[ii] <- "MyFunc"  
    ListIndex <- c(ii, j)
    assign("ListIndex", ListIndex, envir = .GlobalEnv)    
    temp  <- mice.mids(IMtemp, maxit=1)  
    IMinit$imp[listvar[ii]] <- temp$imp[listvar[ii]]
  if(MyMethod[ii]=="polyreg"){
      j <- (j + dim(table(IM$data[ii]))-2)
  }
  cat("\n")
  SumPr[cpt, 1] <- SumPrtemp[1]
  SumPr[cpt, 2] <- SumPrtemp[2]
  SumPr[cpt, 3] <- SumPrtemp[3]
  if(length(SumPrtemp) > 3){
    temp <- SumPrtemp[3]
    for(l in 4 : length(SumPrtemp)){
      temp <- paste(temp, SumPrtemp[l], sep=" ; ")  
    }
    SumPr[cpt, 3] <- temp
  }         
  }
  if(ListMethod[ii]==""){
    IMinit$imp[listvar[ii]] <- IMinit$imp[listvar[ii]]
  }
}
dimnames(SumPr)[[2]] <- c("Variable", "Method", "SupPar") 
cat("Summary :")
cat("\n")
cat("\n")
print(SumPr)
IMfinal <- IMinit
}




##' Internal function for sens.mice
##'
##' @title Internal function for sens.mice
##' @export
##' @keywords internal
mice.impute.MyFunc <- function(y, ry, x, suppar = SupPar, Mymethod = MyMethod, i = ListIndex[1], j = ListIndex[2]){
  if(Mymethod[i]=="pmm"){
    x <- cbind(1, as.matrix(x))
    parm <- .norm.draw(y, ry, x)
    parm$beta[1] <- parm$beta[1] + suppar[j]
    yhatobs <- x[ry, ] %*% parm$coef
    yhatmis <- x[!ry, ] %*% parm$beta
    return(apply(as.array(yhatmis), 1, .pmm.match, yhat = yhatobs,
        y = y[ry]))
  }
  if(Mymethod[i]=="norm"){
    x <- cbind(1, as.matrix(x))
    parm <- .norm.draw(y, ry, x)
    parm$beta[1] <- parm$beta[1] + suppar[j]
    return(x[!ry, ] %*% parm$beta + rnorm(sum(!ry)) * parm$sigma)
  }
  if(Mymethod[i]=="logreg"){
    aug <- mice:::augment(y, ry, x)
    x <- aug$x
    y <- aug$y
    ry <- aug$ry
    w <- aug$w
    x <- cbind(1, as.matrix(x))
    suppressWarnings(fit <- glm.fit(x[ry, ], y[ry], family = binomial(link = logit),
        weights = w[ry]))
    fit.sum <- summary.glm(fit)
    beta <- coef(fit)
    beta[1] <- beta[1] + log(suppar[j])
    rv <- t(chol(fit.sum$cov.unscaled))
    beta.star <- beta + rv %*% rnorm(ncol(rv))
    p <- 1/(1 + exp(-(x[!ry, ] %*% beta.star)))
    vec <- (runif(nrow(p)) <= p)
    vec[vec] <- 1
    if (is.factor(y)) {
        vec <- factor(vec, c(0, 1), levels(y))
    }
    return(vec)
  }
  if(Mymethod[i]=="polyreg"){
    x <- as.matrix(x)
    aug <- mice:::augment(y, ry, x)
    x <- aug$x
    y <- aug$y
    ry <- aug$ry
    w <- aug$w   
    assign("data", cbind.data.frame(y, x), pos = 1)
    fit <- multinom(formula(data), data = data[ry, ], weights = w[ry],
        maxit = 200, trace = FALSE) 
    temp <- matrix(fit$wts, nrow=nlevels(y), byrow=T)  
    for(k in 2:nlevels(y)){
      temp[k, 2] <- temp[k, 2] + log(suppar[j])
      j <- j + 1
    }       
    temp <- t(temp)
    fit$wts <- c(temp[, ])
    post <- predict(fit, data[!ry, ], type = "probs")
    remove("data", pos = 1)
    if (sum(!ry) == 1)
        post <- matrix(post, nr = 1, nc = length(post))
    fy <- as.factor(y)
    nc <- length(levels(fy))
    un <- rep(runif(sum(!ry)), each = nc)
    if (is.vector(post))
        post <- matrix(c(1 - post, post), nc = 2)
    draws <- un > apply(post, 1, cumsum)
    idx <- 1 + apply(draws, 2, sum)
    return(levels(fy)[idx])     
  }
}
